import click
import uvicorn


@click.command()
@click.option(
    "--host", default="127.0.0.1", show_default=True, help="Bind socket to this host."
)
@click.option(
    "--port",
    default=8000,
    show_default=True,
    help="Bind socket to this port. If 0, an available port will be picked.",
)
@click.option("--uds", help="Bind to a UNIX domain socket.")
@click.option(
    "--workers",
    help=(
        "Number of worker processes. "
        "Defaults to the $WEB_CONCURRENCY environment variable if available, or 1."
    ),
    type=int,
)
def cli(**kwargs) -> None:
    uvicorn.run("pssecret_server.main:app", **kwargs)



from pydantic import BaseModel, Field


class Secret(BaseModel):
    data: str = Field(title="Secret", description="Some secret data", min_length=1)


class SecretSaveResult(BaseModel):
    key: str = Field(
        title="Retrieval key",
        description="Key that should be used for retrieval of submitted secret",
    )



from typing import Annotated

from cryptography.fernet import Fernet
from fastapi import Depends

from pssecret_server.settings import Settings, get_settings


def get_fernet(settings: Annotated[Settings, Depends(get_settings)]) -> Fernet:
    return Fernet(settings.secrets_encryption_key)



from pydantic import RedisDsn
from pydantic_settings import BaseSettings, SettingsConfigDict


class Settings(BaseSettings):
    model_config = SettingsConfigDict(env_file=".env", extra="ignore")

    redis_url: RedisDsn = RedisDsn("redis://localhost")
    secrets_encryption_key: bytes


def get_settings() -> Settings:
    return Settings()



from typing import Annotated

from cryptography.fernet import Fernet
from fastapi import Depends, FastAPI
from fastapi.exceptions import HTTPException
from redis.asyncio import Redis

from pssecret_server.fernet import get_fernet
from pssecret_server.models import Secret, SecretSaveResult
from pssecret_server.redis_db import get_redis
from pssecret_server.utils import decrypt_secret, encrypt_secret, save_secret

app = FastAPI()

RedisDep = Annotated[Redis, Depends(get_redis)]
FernetDep = Annotated[Fernet, Depends(get_fernet)]


@app.post(
    "/secret",
    summary="Store secret",
    description=(
        "Submit secret, it is saved on the server, get retrieval key in response. "
        "Use that key to retrieve your data. Key could be used only once, "
        "so use it wisely"
    ),
    response_model=SecretSaveResult,
)
async def set_secret(
    data: Secret, redis: RedisDep, fernet: FernetDep
) -> dict[str, str]:
    data = encrypt_secret(data, fernet)
    return {
        "key": await save_secret(data, redis),
    }


@app.get(
    "/secret/{secret_key}",
    summary="Retrieve secret",
    description=(
        "Returns previously saved data if it is still on the server. "
        "Could be the other way around in two cases: "
        "either it has already been retrieved, either storage timeout has expired"
    ),
    response_model=Secret,
    responses={404: {"description": "The item was not found"}},
)
async def get_secret(
    secret_key: str, redis: RedisDep, fernet: FernetDep
) -> dict[str, bytes]:
    data: bytes | None = await redis.getdel(secret_key)

    if data is None:
        raise HTTPException(404)

    return {
        "data": decrypt_secret(data, fernet),
    }



from functools import lru_cache
from uuid import uuid4

from cryptography.fernet import Fernet
from redis.asyncio import Redis
from redis.exceptions import ResponseError
from redis.typing import ResponseT

from pssecret_server.models import Secret


def encrypt_secret(data: Secret, fernet: Fernet) -> Secret:
    encrypted = fernet.encrypt(data.data.encode()).decode()
    return Secret(data=encrypted)


def decrypt_secret(secret: bytes, fernet: Fernet) -> bytes:
    return fernet.decrypt(secret)


async def get_new_key(redis: Redis) -> str:
    """Returns free Redis key"""
    while True:
        new_key = str(uuid4())

        if not await redis.exists(new_key):
            return new_key


async def save_secret(data: Secret, redis: Redis) -> str:
    """Save passed data, returns retrieval key"""
    new_key = await get_new_key(redis)
    await redis.setex(new_key, 60 * 60 * 24, data.data)

    return new_key


@lru_cache
async def _is_getdel_available(redis: Redis) -> bool:
    """Checks the availability of GETDEL command on the Redis server instance

    GETDEL is not available in Redis prior to version 6.2
    """
    try:
        await redis.getdel("test:getdel:availability")
    except ResponseError:
        return False

    return True


async def getdel(redis: Redis, key: str) -> ResponseT:
    """Gets the value of key and deletes the key

    Depending on the capabilities of Redis server this function
    will either call GETDEL command, either first call GETSET with empty string
    and DEL right after that.
    """
    result: ResponseT

    if await _is_getdel_available(redis):
        result = await redis.getdel(key)
    else:
        result = await redis.getset(key, "")
        await redis.delete(key)

    return result



# noinspection PyUnresolvedReferences,PyProtectedMember
from typing import Annotated

from fastapi import Depends
from redis import asyncio as aioredis

from pssecret_server.settings import Settings, get_settings


def get_redis(settings: Annotated[Settings, Depends(get_settings)]) -> aioredis.Redis:
    return aioredis.from_url(str(settings.redis_url))
